#uses "LHCServices_Config"	 //load dll or so file 

/**
* Parse config file and store data into dyn_strings
*
* @param   sConfigFile     input,  Pathname of the config file
* @param   &dsVariableName output, list of VARIABLE_NAMES from config file  
* @param   &dsDpe          output, list of DPES from config file
* @param   &dsAlias        output, list of ALIASES from config file
* @param   &dsHierarchy    output, list of HIERARCHIES from config file
* @param   &dsDescription  output, list of DESCRIPTIONS from config file
* @param   &dsArchive      output, list of ARCHIVES from config file
* @param   &dsElementId    output, list of ELEMENT_IDS from ELEMENTS table
* @param   &dsUnit         output, list of UNITSS from ELEMENTS table
* @param   &dsRDBSource    output, list of GROUP_NAMES from ELEMENTS table
* @param   &dsErrors       output, list of errors 
* @param   &dsWarnings     output, list of warnings
*
* @return  bSuccess        output, true=success,false=error 
*/
public bool unLHCLoggingDBParser_parseConfigFile(const string sConfigFile, dyn_string &dsVariableName, dyn_string &dsDpe, dyn_string &dsAlias, 
                         dyn_string &dsHierarchy, dyn_string &dsDescription, dyn_string &dsArchive,
                         dyn_string &dsElementId, dyn_string &dsUnit, dyn_string &dsRDBSource, 
                         dyn_string &dsErrors, dyn_string &dsWarnings)
{   
  dynClear(dsVariableName);
  dynClear(dsDpe);
  dynClear(dsAlias);
  dynClear(dsHierarchy);
  dynClear(dsDescription);
  dynClear(dsArchive);
  dynClear(dsElementId);
  dynClear(dsUnit);
  dynClear(dsRDBSource);   
  dynClear(dsErrors);
  dynClear(dsWarnings);   
  
  bool bSuccess=true;   
  
  // Check file format
  if(strpos(strtoupper(sConfigFile), ".XML") >= 0) { 
    // XML-format
    bSuccess &= unLHCLoggingDBParser_parseXmlFile(sConfigFile, dsVariableName, dsDpe, dsAlias, dsHierarchy, dsDescription, dsArchive, dsErrors, dsWarnings); 
  } 
  else { 
    // Text-format
    bSuccess &= unLHCLoggingDBParser_parseTextFile(sConfigFile, dsVariableName, dsDpe, dsAlias, dsHierarchy, dsDescription, dsArchive, dsErrors, dsWarnings);
  }
  bSuccess &= unLHCLoggingDBParser_checkDpConsistency (dsDpe, dsAlias, dsErrors, dsWarnings);
  bSuccess &= unLHCLoggingDBParser_getElementData(dsDpe, dsElementId, dsUnit, dsRDBSource, dsErrors, dsWarnings);
  bSuccess &= unLHCLoggingDBParser_checkMetaVariables(dsDpe, dsElementId, dsVariableName, dsErrors, dsWarnings);
  
  return bSuccess;  
}


/**
* Parse Xml file and store data into dyn_strings
*
* @param   sFileName       input,  Pathname of the XML file
* @param   &dsVariableName output, list of variable names  
* @param   &dsDpe          output, list of dpe names
* @param   &dsAlias        output, list of dpe aliases
* @param   &dsHierarchy    output, list of hierarchies
* @param   &dsDescription  output, list of desciptions
* @param   &dsArchive      output, list of archives
* @param   &dsErrors       output, list of errors 
* @param   &dsWarnings     output, list of warnings
*
* @return                  output, true=success,false=error 
*/
private bool unLHCLoggingDBParser_parseXmlFile(const string sFileName, dyn_string &dsVariableName, dyn_string &dsDpe, dyn_string &dsAlias, 
                      dyn_string &dsHierarchy, dyn_string &dsDescription, dyn_string &dsArchive, 
                      dyn_string &dsErrors, dyn_string &dsWarnings)
{  
  bool returnCode = true;  
  
  int docum, rootNode; 
  dyn_int nodes;
  string errMsg, errLin, errCol;  

  // open XML file
  docum = xmlDocumentFromFile ( sFileName , errMsg , errLin , errCol );  
  if (docum < 0) {
    dynAppend(dsErrors, "Error: Parsing: '" + errMsg + "' Line=" + errLin + " Column=" + errCol);
    return false;
  }
  
  // Find root node
  rootNode = xmlFirstChild (docum);
  if(rootNode == -1) {
    dynAppend(dsErrors, "Error: Root node does not exist");
    return false; 
  }  
  if(xmlNodeType(docum, rootNode) != XML_ELEMENT_NODE) {
    do {
      rootNode = xmlNextSibling(docum, rootNode);
      if(rootNode == -1) {
        dynAppend(dsErrors, "Error: Root node does not exist");
        return false; 
      }
    } while(xmlNodeType(docum, rootNode) != XML_ELEMENT_NODE);
  }
  
  // Retrieve child nodes   
  if(xmlChildNodes(docum, rootNode, nodes) != 0) {
    dynAppend(dsErrors, "Error: While reading child nodes");
    return false;
  }  
  
  // element index
  int id=0;
  
  // Iterate for each node  
  int numberOfNodes = dynlen(nodes); 
  for (int ii=1; ii<=numberOfNodes; ii++)   //dynlen(nodes) is the number of nodes
  {  
    // Animate progress bar    
    progressBar.progress(ii * 25 / numberOfNodes);  // progress from 0 to 25%    
    
    // Only process element nodes
    if ( xmlNodeType(docum, nodes[ii]) == XML_ELEMENT_NODE ) 
    { 
      id++; 
      
      // Get node attribute 'Dpe' (optional)
      unLHCLoggingDBParser_getNodeAttribute(docum, nodes[ii], "Dpe", dsDpe[id]);
    
      // Get node attribute 'Alias' (optional)
      unLHCLoggingDBParser_getNodeAttribute(docum, nodes[ii], "Alias", dsAlias[id]);
  
      // Get node attribute 'Name' (mandatory)
      if(unLHCLoggingDBParser_getNodeAttribute(docum, nodes[ii], "Name", dsVariableName[id]) == false) {
        dynAppend(dsErrors, "Error: Datapoint '" + dsDpe[id] + "' has no attribute: 'Name'");
        returnCode = false;    
      }
 
      // Get node attribute 'Hierarchy' (optional)
      if (unLHCLoggingDBParser_getNodeAttribute(docum, nodes[ii], "Hierarchy", dsHierarchy[id]) == false) {
        dynAppend(dsWarnings, "Warning: Datapoint '" + dsDpe[id] + "' has no attribute: 'Hierarchy'");
      }  
  
      // Get node attribute 'Description' (optional)
      unLHCLoggingDBParser_getNodeAttribute(docum, nodes[ii], "Description", dsDescription[id]);
  
      // Get node attribute 'Archive' (optional)
      unLHCLoggingDBParser_getNodeAttribute(docum, nodes[ii], "Archive", dsArchive[id]);
    }
  } 
  
  if(id==0) {
    returnCode = false;  
  }

  return returnCode;  
}


/**
* Get value of xml node attribute and check if attribute is missing or has no value
*
* @param   doc         input, xml document id
* @param   node        input, node number
* @param   attr        input, attribute name
* @param   &value      output, attribute value
*
* @return              output, true=success,false=error 
*/
private bool unLHCLoggingDBParser_getNodeAttribute (const unsigned doc, const unsigned node, const string attr, string &value) 
{
  if( xmlGetElementAttribute(doc, node, attr, value) == -1 ) 
  { // Attribute not found
    return false;  
  } 
  else 
  { // Attribute was found
    if(value == "") 
    { // Attribute has no value    
      return false;      
    }      
  }  
  return true;  
}


/** 
* Parse text file and store data into dyn_strings
* Piotr: System name will be added to dp
*
* @param   sFileName       input,  Pathname of the XML file
* @param   &dsVariableName output, list of variable names  
* @param   &dsDpe          output, list of dpe names
* @param   &dsAlias        output, list of dpe aliases
* @param   &dsHierarchy    output, list of hierarchies
* @param   &dsDescription  output, list of desciptions
* @param   &dsArchive      output, list of archives
* @param   &dsErrors       output, list of errors 
* @param   &dsWarnings     output, list of warnings
*
* @return                  output, true=success,false=error 
*/
private bool unLHCLoggingDBParser_parseTextFile(const string sFileName,  dyn_string &dsVariableName, dyn_string &dsDpe, dyn_string &dsAlias, 
                       dyn_string &dsHierarchy, dyn_string &dsDescription, dyn_string &dsArchive,
                       dyn_string &dsErrors, dyn_string &dsWarnings )
{ 
 bool returnCode = true;    
  
 int element=0;
	int len, len2;
	dyn_string dsFileLines;
	dyn_dyn_string ddsFile;

	if (isFunctionDefined("ctrlConfRead"))
	{
		ctrlConfRead(sFileName,ddsFile);
		len=dynlen(ddsFile);  

		for (int i=1; i<=len; i++) //read each statement
		{				
			int l=1;
			dsFileLines=ddsFile[i];

			len2=dynlen(dsFileLines);

			progressBar.progress((i*25)/len);  // progress from 0 to 25 %   
   
   string sArchive, sDpe, sAlias, sName, sDescription, sHierarchy; 

			// Get Archive
			if(patternMatch("*_ValueArchive_*", dsFileLines[l]))
				{					
					sArchive=substr(dsFileLines[l],strpos(dsFileLines[l],"_ValueArchive_"));
					sArchive=substr(sArchive,0,strpos(sArchive,"]"));
				}
				else
				{
				l=l+1;
     // Get Dpe
					if(patternMatch("*dpe=*", dsFileLines[l]))
					{      
						sDpe=substr(dsFileLines[l], strpos(dsFileLines[l],"=")+1);

						// Get alias 
						if(patternMatch("*alias=*", dsFileLines[l+1])) 	{
							sAlias=substr(dsFileLines[l+1], strpos(dsFileLines[l+1],"=")+1);
						}
						else {
       dynAppend(dsWarnings,"Warning: Dpe '" + sDpe + "' has no attribute 'alias'");
							continue; //return false;
						}
      
						// Get name 
						if(patternMatch("*name=*", dsFileLines[l+2])) 	{
							sName=substr(dsFileLines[l+2], strpos(dsFileLines[l+2],"=")+1);
       if(sName=="") {
         dynAppend(dsErrors, "Error: Dpe '" + sDpe + "', attribute 'name' has no value");
         returnCode = false;
       }
						}
						else {
       dynAppend(dsErrors, "Error: Dpe '" + sDpe + "' has no attribute 'name'");
       returnCode = false;
							continue; //return false;
						}
												
						// Get description 
						if(patternMatch("*description=*", dsFileLines[l+3])) 	{
							sDescription=substr(dsFileLines[l+3], strpos(dsFileLines[l+3],"=")+1);          
						}
						else	{
       dynAppend(dsWarnings,"Warning: Dpe '" + sDpe + "' has no attribute 'description'");
							continue; //return false;
						}
						
						// Get hierarchy
						if(patternMatch("*hierarchy=*", dsFileLines[l+4]))
						{
							string sNode="";
							string sPattern=dsFileLines[l+4];
							int iPos=strpos(sPattern,"node name=");
							while (iPos>=0)
						 {
						   sPattern=substr(sPattern, iPos + 11, strlen(sPattern) - (iPos+11));
						   if (sNode=="")
						     sNode=substr(sPattern, 0, strlen(sPattern) - (strlen(sPattern) - strpos(sPattern,"\"")));
						   else	
							    sNode=sNode + "/" + substr(sPattern, 0, strlen(sPattern) - (strlen(sPattern) - strpos(sPattern,"\"")));
							  iPos=strpos(sPattern,"node name=");
						 }       
       sHierarchy = sNode;
       strreplace(sHierarchy,"/","->");
       if(sHierarchy=="") {
         dynAppend(dsWarnings,"Warning: Dpe '" + sDpe + "', attribute 'hierarchy' has no value");
       }
						}
						else
						{
       dynAppend(dsWarnings,"Warning: Dpe '" + sDpe + "' has no attribute 'hierarchy'");
							continue; //return false;								
						}
      
      //get format
						if(l+5 <= len2) //line exists
						{
							if(patternMatch("*_ValueArchive_*", dsFileLines[l+5]) || patternMatch("*dpe=*", dsFileLines[l+5])) // no format
							{
							  i=i+1;		
							}
							else // format found
							{}  // don't care
						}
						else //last line of the file reached and format not defined
						{}  // don't care   
      
      element++; 
      
      dsVariableName[element] = sName;
      dsDpe[element] = sDpe;
      dsAlias[element] = sAlias;
      dsHierarchy[element] = sHierarchy;
      dsDescription[element] = sDescription;
      dsArchive[element] = sArchive;       

					} // pattern dpe
					else
					{
      dynAppend(dsErrors, "Error: Archive " + sArchive + " has no attribute 'dpe'");
      returnCode = false;
      continue; //return false;
					}	
				}		// pattern archive
			}			// for
		}					
		else
		{
		 DebugN("************* DLL not defined *************");   
   return false;
		}  
  
  return returnCode; 
}


/**
* Check consistency between datapoint and alias
*
* @param   &dsDpe          input/output, list of dpe names
* @param   &dsAlias        input/output, list of dpe aliases
* @param   &dsErrors       output, list of errors 
* @param   &dsWarnings     output, list of warnings
*
* @return                  output, true=success,false=error 
*/
private bool unLHCLoggingDBParser_checkDpConsistency (dyn_string &dsDpe, dyn_string &dsAlias, dyn_string &dsErrors, dyn_string &dsWarnings)
{ 
  bool returnValue=true;   
  
  for(int i=1;i<=dynlen(dsDpe);i++)  
  { 
    progressBar.progress(25 + i * 25 / dynlen(dsDpe));  // progress from 25 to 50 %        
   
    // Get datapoint name including SystemName (e.g.: dist_1:...) 
    string sDpeWithSystemName = dpSubStr(dsDpe[i],DPSUB_SYS_DP_EL); // TOTEST
    if (sDpeWithSystemName != "") { 
      // If there is a system name, add it to dsDpe, if not, leave dsDpe as it is
      dsDpe[i] = sDpeWithSystemName;
    }           
 
     // If there is no dp and no alias
    if((dsDpe[i] == "") && (dsAlias[i] == "")) 
    {  
      dynAppend(dsErrors, "Error: Element " + i + ", neither datapoint nor alias specified");
      returnValue = false;  
    } 
 
     // If there is a dp and no alias
    else if((dsDpe[i] != "") && (dsAlias[i] == "")) 
    {  
      if(dpExists(dsDpe[i]) == false) 
      {   
        dynAppend(dsErrors, "Error: Datapoint '" + dsDpe[i] + "' does not exist on curent system");
        dsDpe[i] = "";
        returnValue = false;   
      }
    } 
 
     // If there is no dp but an alias   
    else if((dsDpe[i] == "") && (dsAlias [i]!= "")) 
    {    
      // Retrieve alias to dp
      string sTempDpe = dpAliasToName(dsAlias[i]);
 
      //If dp does not exist and if di DPNAME    
      if(dpExists(sTempDpe) == false)
      {
        dynAppend(dsErrors, "Error: Alias: '" + dsAlias[i] + "' does not exist");  
        dsDpe[i] = "";
        returnValue = false;   
      }
      else if(unLHCLoggingDBParser_isDpName(sTempDpe) == true) 
      {
        dynAppend(dsErrors, "Error: Alias: '" + dsAlias[i] + "' belongs to the DPName:'"+ sTempDpe +"', Please specify the Dpe field");
        dsDpe[i] = "";
        returnValue = false;   
      }
      else
      {
        // Prepare real dpe name for registartion
        dsDpe[i] = sTempDpe;
      }      
    }    
    
    // If there is a datapoint name and an alias  
    else if((dsDpe[i] != "") && (dsAlias[i] != "")) 
    {       
      // Retrieve alias to dp
      string sTempDpe = dpAliasToName(dsAlias[i]);    
    
      // Check if Alias is a DPname or a DPelement. It's a DPname, if the last char is a '.'
      if(unLHCLoggingDBParser_isDpName(sTempDpe) == true)
      {
        // Alias is a DPNAME 
        if(!dpExists(sTempDpe + dsDpe[i]))  // Check if Alias + dsDpe is a valid DPE
        {
          dynAppend(dsErrors, "Error: Wrong Alias or DPe specified. Alias: '" + dsAlias[i] + "' resolved to DPName: '" + sTempDpe + "' and Dpe: '" + dsDpe[i] + "'");
          dsDpe[i] = "";
          returnValue = false;   
        }
        else
        {
          // Prepare real dpe name for registartion
          dsDpe[i] = sTempDpe + dsDpe[i]; 
        }
      }
      else
      {
        // Alias is a DPE
        if((dsDpe[i] != sTempDpe) || !dpExists(sTempDpe))   
        {
          dynAppend(dsErrors, "Error: Alias: '" + dsAlias[i] + "' and Dpe: '" + dsDpe[i] + "' do not match");
          dsDpe[i] = "";
          returnValue = false;   
        }
        else
        {
          // Leave dsDpe[i] as it is because its correct
        }          
      }   
    }    
  }  
  return returnValue;
}


/**
* Checks if a string is a DP name 
*
* @param   &sDp          input, DP string
* @return                output, true: it's a DP name, false: it's no DP name
*/
private bool unLHCLoggingDBParser_isDpName(string sDp)
{
  int length = strlen(sDp);
  
  if(strpos(sDp, ".", length-1) >= 0) {
    // Dot was found at last position -> it is a DP name
    return true; 
  }  
  return false; 
}


/**
* Get additional data from ELEMENTS table
*
* @param   &dsDpe          input, list of DPES
* @param   &dsElementId    output, list of ELEMENT_IDS from ELEMENTS table
* @param   &dsUnit         output, list of UNITSS from ELEMENTS table
* @param   &dsRDBSource    output, list of RDBSources from ELEMENTS table
* @param   &dsErrors       output, list of errors 
* @param   &dsWarnings     output, list of warnings
*
* @return                  output, true=success,false=error 
*/
public bool unLHCLoggingDBParser_getElementData(const dyn_string &dsDpe, dyn_string &dsElementId, dyn_string &dsUnit, dyn_string &dsRDBSource, dyn_string &dsErrors, dyn_string &dsWarnings, bool bEnableProgressBar=true)
{
  bool ret=true;  
  
  string sMySystemDbSchema = unLHCLoggingDBSettings_getMySystemDbSchema();
  
  string sqlQuery =  "select TO_CHAR(ELEMENT_ID), UNIT, GROUP_NAME from " + sMySystemDbSchema + ".ELEMENTS where ELEMENT_NAME = :dpe";
                            
  string errTxt;
  dbCommand cmd;
  
  fwDbStartCommand(g_unLHCLoggingDBSql_dbConnection, sqlQuery, cmd);
  if (fwDbCheckError(errTxt,cmd)){DebugN("DB ERROR",errTxt);return false;};      
  
  // derive 10 rows of data, and insert them using the statement above:
  for (int i=1;i<=dynlen(dsDpe);i++) 
  { 
    if (bEnableProgressBar==true) progressBar.progress(50 + i * 25 / dynlen(dsDpe));  // progress from 50 to 75 %      
    
    dsElementId[i] = "";  
    dsUnit[i] = ""; 
    dsRDBSource[i] = "";  
    
    if(dsDpe[i]=="") continue;       
    
    mapping record;
    record["dpe"] = dsDpe[i];

    fwDbBindParams(cmd, record);
    if (fwDbCheckError(errTxt,cmd)){DebugN("DB ERROR",errTxt);return false;};

    fwDbExecuteCommand(cmd);
    if (fwDbCheckError(errTxt,cmd)){DebugN("DB ERROR",errTxt);return false;};
      
    dyn_dyn_mixed sqlResult;
    fwDbGetData(cmd, sqlResult, false);
    if (fwDbCheckError(errTxt,cmd)){DebugN("DB ERROR",errTxt);return false;};
      
    if(dynlen(sqlResult) > 0) {
      dsElementId[i] = sqlResult[1][1];  
      dsUnit[i] = sqlResult[1][2]; 
      dsRDBSource[i] = sMySystemDbSchema + "." + sqlResult[1][3] + "HISTORY"; 
    }
    else { // Dpe does not exist in .ELEMENTS table 
      dynAppend(dsErrors, "Error: Datapoint '" + dsDpe[i] + "' not found in ELEMENTS table "); 
      ret=false;        
    }
  };

  fwDbFinishCommand(cmd);
  if (fwDbCheckError(errTxt)){DebugN("DB ERROR",errTxt);return false;}; 
  
  return ret;
}


/**
* Check if list of variables is already registered in META_VARIABLES table 
*
* @param   &dsDpe          input, list of DPES
* @param   &dsElementId    input, list of ELEMENT_IDS from ELEMENTS table
* @param   &dsVariableName input, list of VARIABLE_NAMES from config file
* @param   &dsErrors       output, list of errors 
* @param   &dsWarnings     output, list of warnings
*
* @return                  output, true = no variable already registered, false = at least 1 variable already registered 
*/
private bool unLHCLoggingDBParser_checkMetaVariables(const dyn_string &dsDpe, const dyn_string &dsElementId, const dyn_string &dsVariableName, dyn_string &dsErrors, dyn_string &dsWarnings)
{
  bool ret=true;  
  
  string sqlQuery = "select VARIABLE_ID from RDB2LHC.META_VARIABLES where ELEMENT_ID=:elementId or VARIABLE_NAME=upper(:variableName)";
                            
  string errTxt;
  dbCommand cmd;
  
  fwDbStartCommand(g_unLHCLoggingDBSql_dbConnection, sqlQuery, cmd);
  if (fwDbCheckError(errTxt,cmd)){DebugN("DB ERROR",errTxt);return false;};      
  
  // derive 10 rows of data, and insert them using the statement above:
  for (int i=1;i<=dynlen(dsElementId);i++) 
  {
    progressBar.progress(75 + i * 25 / dynlen(dsElementId)); // progress from 75 to 100 %       
    
    if(dsElementId[i]=="" || dsVariableName[i]=="") continue;       
    
    mapping record;
    record["elementId"] = dsElementId[i];
    record["variableName"] = dsVariableName[i];

    fwDbBindParams(cmd, record);
    if (fwDbCheckError(errTxt,cmd)){DebugN("DB ERROR",errTxt);return false;};

    fwDbExecuteCommand(cmd);
    if (fwDbCheckError(errTxt,cmd)){DebugN("DB ERROR",errTxt);return false;};
      
    dyn_dyn_mixed sqlResult;
    fwDbGetData(cmd, sqlResult, false);
    if (fwDbCheckError(errTxt,cmd)){DebugN("DB ERROR",errTxt);return false;};
      
    if(dynlen(sqlResult) > 0) {
      dynAppend(dsErrors, "Error: Datapoint '" + dsDpe[i]  + "' or Name '" + dsVariableName[i]  + "' already registered in META_VARIABLES table ");
      ret=false;
    }
  };

  fwDbFinishCommand(cmd);
  if (fwDbCheckError(errTxt)){DebugN("DB ERROR",errTxt);return false;}; 
  
  return ret;
}

/**
* Create new config file for logging configuration in xml format
*
* @param   sFilePathName   input, full pathname of the file
* @param   &dsDpe          input, list of dpe names
* @param   &dsAlias        input, list of dpe aliases
* @param   &dsName         input, list of variable names  
* @param   &dsHierarchy    input, list of hierarchies
* @param   &dsDescription  input, list of desciptions
*
* @return                  output, true=success, false=error
*/
public bool unLHCLoggingDBParser_createConfigFile(string sFilePathName, 
                    const dyn_string &dsDpe, const dyn_string &dsAlias, const dyn_string &dsName, 
                    const dyn_string &dsHierarchy, const dyn_string &dsDescription)
{
  int docNum, rootNode, childNode;
  
  docNum = xmlNewDocument();
  xmlAppendChild(docNum, -1, XML_PROCESSING_INSTRUCTION_NODE, "xml version=\"1.0\"");  
  
  rootNode = xmlAppendChild(docNum, -1, XML_ELEMENT_NODE, "unLHCLoggingDB:ConfigFile"); 
  for (int i=1; i<=dynlen(dsDpe); i++) 
  {
    childNode = xmlAppendChild(docNum, rootNode, XML_ELEMENT_NODE, "Element");
    xmlSetElementAttribute(docNum, childNode, "Dpe", dsDpe[i]);
    xmlSetElementAttribute(docNum, childNode, "Alias", dsAlias[i]);
    xmlSetElementAttribute(docNum, childNode, "Name", dsName[i]);
    xmlSetElementAttribute(docNum, childNode, "Hierarchy", dsHierarchy[i]);
    xmlSetElementAttribute(docNum, childNode, "Description", dsDescription[i]);
  }
  
  xmlDocumentToFile(docNum, sFilePathName);
  
  return true;
}
