// Datapoint that holds the name of the active configuration datapoint 
const string sSettingsDP = "unLHCLoggingDB_Settings";

// Datapoints that hold database settings
const string sDbConfigDpType = "_UnLHCLoggingDB_Database";
const string sDbConfigDpPrefix = "unLHCLoggingDB_Database_"; 


public string unLHCLoggingDBSettings_createDbConfigDpName(string sDbName) {
 return (getSystemName() + sDbConfigDpPrefix + sDbName); 
}

/**
* Get name of settings dpe 
*
* @param sDpeName     output, Name of the dpe that holds the settings
* @return             output, true: success, false: error
*/
public bool unLHCLoggingDBSettings_getSettingsDpeName(string &sDpeName)
{  
  sDpeName = getSystemName() + sSettingsDP + ".activeDatabase";  
  if(dpExists(sDpeName) == true) {
    return true;
  }  
  return false;
}


/**
* Get name of datapoint, that holds database settings
*
* @param sDpName      output, Name of the database configuration datapoint
* @return             output, true: success, false: error
*/
public bool unLHCLoggingDBSettings_getActiveDbConfigDp(string &sDpName)
{   
  string sSettingsDpe;  
  if(unLHCLoggingDBSettings_getSettingsDpeName(sSettingsDpe) == true) {
    dpGet(sSettingsDpe, sDpName);   
    return true; 
  }  
  return false;  
}


/**
* Set name of datapoint, that holds database settings
*
* @param sDpName      input, Name of the database configuration datapoint
* @return             output, true: success, false: error
*/
public bool unLHCLoggingDBSettings_setActiveDbConfigDp(string sDpName)
{
  string sSettingsDpe;  
  if(unLHCLoggingDBSettings_getSettingsDpeName(sSettingsDpe) == true) {
    dpSet(sSettingsDpe, sDpName);  
    return true;  
  }  
  return false;   
}


/**
* Get data from a specific database configuration datapoint 
*
* @param sDbConfigDp    input, Name of the database configuration datapoint
* @param sDbName        output, database name 
* @param sDbUser        output, database user 
* @param sDbPass        output, database password (encrypted)
* @return               output, true: success, false: error
*/
public bool unLHCLoggingDBSettings_getDbConfig(string sDbConfigDp, string &sDbName, string &sDbUser, string &sDbPass)
{ 
  if(dpExists(sDbConfigDp))
  { 
    dpGet(sDbConfigDp + ".dbName", sDbName);  
    dpGet(sDbConfigDp + ".dbUser", sDbUser);          
    dpGet(sDbConfigDp + ".dbPass", sDbPass);   
    return true;
  }
  return false;
}

/**
* Set data of a specific database configuration datapoint 
*
* @param sDbConfigDp    input, Name of the database configuration datapoint
* @param sDbName        input, database name 
* @param sDbUser        input, database user 
* @param sDbPass        input, database password (encrypted)
* @return               output, true: success, false: error
*/
private bool unLHCLoggingDBSettings_setDbConfig(string sDbConfigDp, string sDbName, string sDbUser, string sDbPass)
{ 
  if(dpExists(sDbConfigDp))
  { 
    dpSet(sDbConfigDp + ".dbName", sDbName);  
    dpSet(sDbConfigDp + ".dbUser", sDbUser);          
    dpSet(sDbConfigDp + ".dbPass", sDbPass);   
    return true;
  }
  return false;
}


/**
* Get database settings of active database
*
* @param sDbName        output, database name 
* @param sDbUser        output, database user 
* @param sDbPass        output, database password (encrypted) 
* @return               output, true: success, false: error
*/
public bool unLHCLoggingDBSettings_getActiveDbConfig(string &sDbName, string &sDbUser, string &sDbPass)
{
  string sDbConfigDp;  

  if(unLHCLoggingDBSettings_getActiveDbConfigDp(sDbConfigDp) == true)
  {
    if(unLHCLoggingDBSettings_getDbConfig(sDbConfigDp, sDbName, sDbUser, sDbPass) == true)
    {
      return true;
    }
  }
  return false; 
}


/**
* Get database settings of all database configuration datapoints
*
* @param dsDbConfigDps    output, database configuration datapoint names
* @param dsDbNames        output, database names 
* @param dsDbUsers        output, database users 
* @param dsDbPasswords    output, database passwords (encrypted) 
* @param dbDpEditable     output, true=dbConfigDp is editable; false dbConfigDp is read only 
* @return                 output, true: success, false: error
*/
public bool unLHCLoggingDBSettings_getAllDbConfigDps(dyn_string &dsDbConfigDps, dyn_string &dsDbNames, dyn_string &dsDbUsers, dyn_string &dsDbPasswords, dyn_bool &dbDpEditable)
{
  dynClear(dsDbConfigDps);
  dynClear(dsDbNames);
  dynClear(dsDbUsers);
  dynClear(dsDbPasswords);
  dynClear(dbDpEditable);
  
  dsDbConfigDps = dpNames("*","_UnLHCLoggingDB_Database");    
  
  string sSystemName = getSystemName();   
  
  for(int i=1;i<=dynlen(dsDbConfigDps);i++) 
  {
    // Only dbConfigDps of current system are editable
    dbDpEditable[i] = patternMatch("*"+sSystemName+"*", dsDbConfigDps[i]);    
    unLHCLoggingDBSettings_getDbConfig(dsDbConfigDps[i], dsDbNames[i], dsDbUsers[i], dsDbPasswords[i]);   
  }

  return true;  
}


/**
* Set database settings of all database configuration datapoints
*
* @param dsDbConfigDps    input, database configuration datapoint names
* @param dsDbNames        input, database names 
* @param dsDbUsers        input, database users 
* @param dsDbPasswords    input, database passwords (encrypted) 
* @return                 output, true: success, false: error
*/
public bool unLHCLoggingDBSettings_setAllDbConfigDps(const dyn_string &dsDbConfigDps, const dyn_string &dsDbNames, const dyn_string &dsDbUsers, const dyn_string &dsDbPasswords)
{
  // Get SystemName and SystemId
  string sSystemName = getSystemName(); 
  int iSystemId = getSystemId();    

  dyn_string dsDbConfigDpsOld = dpNames("*","_UnLHCLoggingDB_Database");  

  // Compare lists of old dbConfigDps with new ones and delete DPs 
  // that are no more in the new list (because they have been deleted by the user)
  for (int i=1; i<=dynlen(dsDbConfigDpsOld); i++) {
    if(dynContains(dsDbConfigDps, dsDbConfigDpsOld[i]) == 0) {
       if(patternMatch("*" + sSystemName + "*", dsDbConfigDpsOld[i]) == true) {
        // Only datapoints of this system can be deleted
        dpDelete(dsDbConfigDpsOld[i]); 
      } 
    }
  }   
  
  // Compare lists of new dbConfigDps with old ones and create DPs 
  // that are not in the old list (because they have been added by the user)
  for(int i=1; i<=dynlen(dsDbConfigDps);i++) 
  { 
    // Check if new element already exists
    int iIndex = dynContains(dsDbConfigDpsOld, dsDbConfigDps[i]);
    if (iIndex > 0)
    { 
      // Update values of datapoint 
      if(patternMatch("*" + sSystemName + "*", dsDbConfigDps[i]) == true) 
      {
        // Only datapoints of this system can be changed
        unLHCLoggingDBSettings_setDbConfig(dsDbConfigDps[i], dsDbNames[i], dsDbUsers[i], dsDbPasswords[i]);
      } 
    }  
    else if(iIndex == 0) 
    {
      dyn_string dpName = strsplit(dsDbConfigDps[i], ":");  // TODO: secure it, what happes when there is no ":"    
      // A new datapoint has to be created (on this system) because it was added by the user
      dpCreate(dpName[2],sDbConfigDpType,  iSystemId);   
      unLHCLoggingDBSettings_setDbConfig(dsDbConfigDps[i], dsDbNames[i], dsDbUsers[i], dsDbPasswords[i]);
    }  
  } 
  
  return true; 
} 
  
  
/**
* Get db user from the archive configuration of the current system
*
* @return   mySystemDbSchema     output, Schema name as string         
*/
public string unLHCLoggingDBSettings_getMySystemDbSchema()
{
  string sSchema;
  dpGet("_RDBArchive.db.user", sSchema);
  return sSchema;
}

/**
* Read mode of time stamps from datapoint
*
* @return   bUTC       output, 0=localtime 1=UTC       
*/
public bool unLHCLoggingDBSettings_getTimeStampMode()
{
  bool bUTC;
  dpGet(getSystemName() + sSettingsDP + ".timeAsUTC", bUTC);
  return bUTC;  
}

/**
* Write mode of time stamps to datapoint
*
* @return   bUTC       input, 0=localtime 1=UTC       
*/
public void unLHCLoggingDBSettings_setTimeStampMode(bool bUTC)
{  
  dpSet(getSystemName() + sSettingsDP + ".timeAsUTC", bUTC);
}

